/**
 * Open Command Prompt Shell Extension
 * Copyright (C) 2007 Kai Liu.  All rights reserved.
 *
 * Please refer to readme.txt for information about this source code.
 * Please refer to license.txt for details about distribution and modification.
 **/

#include "CCmdOpen.hpp"

STDMETHODIMP CCmdOpen::QueryInterface( REFIID riid, LPVOID *ppv )
{
	*ppv = NULL;

	if (IsEqualIID(riid, IID_IUnknown))
	{
		*ppv = this;
	}
	else if (IsEqualIID(riid, IID_IShellExtInit))
	{
		*ppv = (LPSHELLEXTINIT)this;
	}
	else if (IsEqualIID(riid, IID_IContextMenu))
	{
		*ppv = (LPCONTEXTMENU)this;
	}

	if (*ppv)
	{
		AddRef();
		return(S_OK);
	}
	else return(E_NOINTERFACE);
}

STDMETHODIMP CCmdOpen::Initialize( LPCITEMIDLIST pidlFolder, LPDATAOBJECT pdtobj, HKEY hkeyProgID )
{
	// Check the ContextMenuHandlers keys to make sure that we're working with
	// "our" handlers, because the shell will call this function multiple times
	// with various handlers, so we should cull the redundant calls.
	if (!hkeyProgID) return(E_INVALIDARG);

	HKEY hkeyHandler;
	TCHAR szBuffer[64];
	wsprintf(szBuffer, TEXT("ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen);

	if (RegOpenKeyEx(hkeyProgID, szBuffer, 0, KEY_QUERY_VALUE, &hkeyHandler) == ERROR_SUCCESS)
	{
		DWORD dwLength = sizeof(szBuffer);
		DWORD dwType;

		if ( RegQueryValueEx(hkeyHandler, TEXT("Type"), NULL, &dwType, (LPBYTE)szBuffer, &dwLength) == ERROR_SUCCESS &&
		     dwType == REG_SZ && dwLength > 2 )
		{
			// The registry does not check for proper null termination, so in order
			// to prevent segfaults in the case of a malformed registry entry, make
			// sure that, at the very least, the final character is a null
			szBuffer[countof(szBuffer) - 1] = 0;

		} else szBuffer[0] = 0;

		RegCloseKey(hkeyHandler);
	}
	else return(E_INVALIDARG);


	// Everybody needs a buffer; and let's double it just to be safe
	TCHAR szPath[MAX_PATH << 1] = { 0 };

	// Clear the list, so that if any of these fail, the list will be empty
	m_lsszPaths.clear();

	// Background: just translate the folder ID
	if (lstrcmp(szBuffer, CLSTYPE_STR_CmdOpenBG) == 0)
	{
		if (pidlFolder && SHGetPathFromIDList(pidlFolder, szPath))
			m_lsszPaths.push_back(szPath);
	}

	// My Documents: just look up the folder path
	else if (lstrcmp(szBuffer, CLSTYPE_STR_CmdOpenMyDocs) == 0)
	{
		if (SHGetSpecialFolderPath(NULL, szPath, CSIDL_PERSONAL, FALSE))
			m_lsszPaths.push_back(szPath);
	}

	// Normal: convert pdtobj to hDrop and then enumerate
	else if (lstrcmp(szBuffer, CLSTYPE_STR_CmdOpenFolder) == 0)
	{
		FORMATETC etc = { CF_HDROP, NULL, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
		STGMEDIUM stg = { TYMED_HGLOBAL };
		HDROP hDrop;

		if (!pdtobj || pdtobj->GetData(&etc, &stg) != S_OK)
			return(E_INVALIDARG);

		if (hDrop = (HDROP)GlobalLock(stg.hGlobal))
		{
			UINT uDrops = DragQueryFile(hDrop, 0xFFFFFFFF, NULL, 0);

			for (UINT uDrop = 0; uDrop < uDrops; ++uDrop)
			{
				if ( DragQueryFile(hDrop, uDrop, szPath, countof(szPath)) &&
				     GetFileAttributes(szPath) & FILE_ATTRIBUTE_DIRECTORY )
				{
					m_lsszPaths.push_back(szPath);
				}
			}

			GlobalUnlock(stg.hGlobal);
		}

		ReleaseStgMedium(&stg);
	}


	// If there was any failure, the list would be empty...
	return((m_lsszPaths.size() > 0) ? S_OK : E_INVALIDARG);
}

STDMETHODIMP CCmdOpen::QueryContextMenu( HMENU hmenu, UINT indexMenu, UINT idCmdFirst, UINT idCmdLast, UINT uFlags )
{
	if (uFlags & CMF_DEFAULTONLY || uFlags & CMF_NOVERBS)
		return(MAKE_HRESULT(SEVERITY_SUCCESS, FACILITY_NULL, 0));

	// Load the localized menu text
	TCHAR szMenuText[48] = { 0 };
	LoadString(g_hModThisDll, IDS_MENUTEXT, szMenuText, countof(szMenuText));

	if (InsertMenu(hmenu, indexMenu, MF_STRING | MF_BYPOSITION, idCmdFirst, szMenuText))
		return(MAKE_HRESULT(SEVERITY_SUCCESS, FACILITY_NULL, 1));

	return(MAKE_HRESULT(SEVERITY_SUCCESS, FACILITY_NULL, 0));
}

STDMETHODIMP CCmdOpen::InvokeCommand( LPCMINVOKECOMMANDINFO pici )
{
	// Ignore string verbs
	if (HIWORD(pici->lpVerb) != NULL)
		return(E_INVALIDARG);

	// The only valid command index is 0
	if (LOWORD(pici->lpVerb) != 0)
		return(E_INVALIDARG);

	// Get the path to cmd.exe
	TCHAR szCommand[MAX_PATH + 0x10] = { TEXT('"') };
	UINT uSize = GetSystemDirectory(szCommand + 1, MAX_PATH + 1);
	if (uSize && uSize < MAX_PATH)
		lstrcat(szCommand, TEXT("\\cmd.exe\""));
	else
		lstrcpyn(szCommand, TEXT("cmd.exe"), countof(szCommand));

	// Load the localized command prompt title
	TCHAR szCmdTitle[32] = { 0 };
	LoadString(g_hModThisDll, IDS_CMDTITLE, szCmdTitle, countof(szCmdTitle));

	// Initialize startup info
	PROCESS_INFORMATION pi;
	STARTUPINFO si;

	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	si.lpTitle = szCmdTitle;

	// Now for the trivial part...
	LPCTSTR lpPath;

	for ( ITR_LSTTSTR it = m_lsszPaths.begin();
	      it != m_lsszPaths.end() && (lpPath = it->c_str());
	      ++it )
	{
		BOOL bSuccess = CreateProcess(
			NULL,
			szCommand,
			NULL,
			NULL,
			FALSE,
			CREATE_DEFAULT_ERROR_MODE,
			NULL,
			lpPath,
			&si,
			&pi
		);

		if (bSuccess)
		{
			CloseHandle(pi.hProcess);
			CloseHandle(pi.hThread);
		}
	}

	return(S_OK);
}

STDMETHODIMP CCmdOpen::GetCommandString( UINT_PTR idCmd, UINT uFlags, UINT *pwReserved, LPSTR pszName, UINT cchMax )
{
	if (idCmd != 0)
		return(E_INVALIDARG);

	HRESULT hr = E_INVALIDARG;

	switch (uFlags)
	{
		case GCS_HELPTEXTA:
			LoadStringA(g_hModThisDll, IDS_MENUTEXT, (LPSTR)pszName, cchMax);
			hr = S_OK;
			break;

		case GCS_HELPTEXTW:
			LoadStringW(g_hModThisDll, IDS_MENUTEXT, (LPWSTR)pszName, cchMax);
			hr = S_OK;
			break;

		// In Vista, the build-in command has a verb of "cmd", so if we match
		// Vista's verb, then Vista's built-in facility will override ours when
		// it is active (this is the desired behavior).

		case GCS_VERBA:
			lstrcpynA((LPSTR)pszName, "cmd", cchMax);
			hr = S_OK;
			break;

		case GCS_VERBW:
			lstrcpynW((LPWSTR)pszName, L"cmd", cchMax);
			hr = S_OK;
			break;
	}

	return(hr);
}
